﻿using System;
using UnityEngine;
using System.Collections.Generic;

namespace ky1001
{
    public class L_R_Mirror: MVRScript 
    {
        #region PluginInfo    
        public string pluginAuthor = "ky1001";
        public string pluginName = "L-R_Mirror";
        public string pluginVersion = "2.0";
        public string pluginDate = "2022-02-22";
        public string pluginDescription = "Renamed from SimpleMirroring.";
        #endregion

        #region Vars
        protected Atom CurrentAtom;
        private Vector3 tempPosition;
        private Vector3 temprPosition;
        private Vector3 templPosition;
        private Quaternion tempRotation;
        private Quaternion temprRotation;
        private Quaternion templRotation;
        private Vector3 tempPosition_h;
        public string tempPosi_r;
        public string tempRota_r;
        public string tempPosi_l;
        public string tempRota_l;
        public string HorF, CF, CF1, CT, CT1;
        protected UIDynamicButton rHbutton;
        protected UIDynamicButton rFbutton;
        protected UIDynamicButton lHbutton;
        protected UIDynamicButton lFbutton;
        protected UIDynamicButton ubutton;
        protected JSONStorableBool toggle;
        Dictionary<string, string> PosiDic = new Dictionary<string, string>();
        Dictionary<string, string> RotaDic = new Dictionary<string, string>();
        Dictionary<string, string> AtomDic = new Dictionary<string, string>();
        Dictionary<string, string> LinkDic = new Dictionary<string, string>();
        Dictionary<string, FreeControllerV3.PositionState> PosiST = new Dictionary<string, FreeControllerV3.PositionState>();
        Dictionary<string, FreeControllerV3.RotationState> RotaST = new Dictionary<string, FreeControllerV3.RotationState>();
        Dictionary<string, Rigidbody> RigidbodyList = new Dictionary<string, Rigidbody>();
        Dictionary<string, float> HandPose = new Dictionary<string, float>();
        Dictionary<string, Vector3> PosiV3 = new  Dictionary<string, Vector3>();
        Dictionary<string, Quaternion> RotaV3 = new  Dictionary<string, Quaternion>();        
        public string RigidbodyName;
        Rigidbody linkToAtom;

        protected JSONStorableStringChooser atomJSON;
        protected UIDynamicPopup dp;
        private UIDynamicButton rfl_button;
        private bool canUndo = false;
        #endregion

        public readonly string[] ControlNames = new string[] 
        {
            "chestControl",
            "headControl",
            "neckControl",
            "pelvisControl",
            "hipControl",
            "abdomenControl",
            "abdomen2Control",
            "eyeTargetControl",
            "penisBaseControl",
            "penisMidControl",
            "penisTipControl",
        };

        public readonly string[] lrControlNames = new string[] 
        {
            "HandControl",
            "FootControl",
            "NippleControl",
            "ElbowControl",
            "KneeControl",
            "ToeControl",
            "ThighControl",
            "ArmControl",
            "ShoulderControl",
        };

        public readonly string[] lrHandControlNames = new string[] 
        {
            "HandControl",
            "ElbowControl",
            "ArmControl",
            "ShoulderControl",
        };

        public readonly string[] lrFootControlNames = new string[] 
        {
            "FootControl",
            "KneeControl",
            "ToeControl",
            "ThighControl",
        };

        private readonly string[] OnControlNames = new string[] 
        {
            "chestControl",
            "headControl",
            "hipControl",
            "rHandControl",
            "lHandControl",
            "rFootControl",
            "lFootControl"
        };

        private readonly string[] lrPoseNames = new string[] 
        {
            "Hand Chop",
            "Hand Fist",
            "Hand Grasp",
            "Hand Spread",
            "Hand Straighten",
            "Fingers Fist",
            "Fingers Grasp",
            "Fingers In-Out",
            "Fingers Squeeze",
            "Fingers Straighten",
            "Index Finger Bend",
            "Mid Finger Bend",
            "Pinky Finger Bend",
            "Ring Finger Bend",
            "Thumb Bend",
            "Thumb Fist",
            "Thumb Grasp",
            "Thumb In-Out",
            "Thumb Straighten",
            "Big Toe Bend",
            "Index Toe Bend",
            "Mid Toe Bend",
            "Pinky Toe Bend",
            "Ring Toe Bend",
            "Toes Fan Down",
            "Toes Small Curl",
            "Toes Spread",
        };

        private readonly string[] lrHandPoseNames = new string[] 
        {
            "Hand Chop",
            "Hand Fist",
            "Hand Grasp",
            "Hand Spread",
            "Hand Straighten",
            "Fingers Fist",
            "Fingers Grasp",
            "Fingers In-Out",
            "Fingers Squeeze",
            "Fingers Straighten",
            "Index Finger Bend",
            "Mid Finger Bend",
            "Pinky Finger Bend",
            "Ring Finger Bend",
            "Thumb Bend",
            "Thumb Fist",
            "Thumb Grasp",
            "Thumb In-Out",
            "Thumb Straighten",
        };

        private readonly string[] lrFootPoseNames = new string[] 
        {
            "Big Toe Bend",
            "Index Toe Bend",
            "Mid Toe Bend",
            "Pinky Toe Bend",
            "Ring Toe Bend",
            "Toes Fan Down",
            "Toes Small Curl",
            "Toes Spread",
        };

	    // Use this for initialization
        public override void Init()
        {
            try
            {
                #region Buttons
                atomJSON = new JSONStorableStringChooser("Target", null, "None", "Target", clearUndo);
                RegisterStringChooser(atomJSON);
                dp = CreateScrollablePopup(atomJSON, false);
                dp.popupPanelHeight = 650f;
                dp.popup.selectColor = new Color(0.8f, 0.8f, 0.8f);

                mkList();

                var spacer = CreateSpacer(true);
                spacer.height = 30f; 

                rfl_button = CreateButton("Refresh List", true);
                if (rfl_button != null) {
                    rfl_button.button.onClick.AddListener(mkList);
                }
                JSONStorableAction refreshListAct = new JSONStorableAction("Refresh List", () =>{mkList();});					
                RegisterAction(refreshListAct);

                spacer = CreateSpacer(true);
                spacer.height = 35f; 

                spacer = CreateSpacer(false);
                spacer.height = 30f; 

                lHbutton = CreateButton("Hand : Left to Right");
                if (lHbutton != null)
                {
                    lHbutton.button.onClick.AddListener(CopyHandFromLeftToRight);
                }
                JSONStorableAction lHbuttonAct = new JSONStorableAction("Hand:Left to Right", () =>{CopyHandFromLeftToRight();});					
                RegisterAction(lHbuttonAct);

                lFbutton = CreateButton("Foot : Left to Right");
                if (lFbutton != null)
                {
                    lFbutton.button.onClick.AddListener(CopyFootFromLeftToRight);
                }
                JSONStorableAction lFbuttonAct = new JSONStorableAction("Foot:Left to Right", () =>{CopyFootFromLeftToRight();});					
                RegisterAction(lFbuttonAct);

                rHbutton = CreateButton("Hand : Right to Left", true);
                if (rHbutton != null)
                {
                    rHbutton.button.onClick.AddListener(CopyHandFromRightToLeft);
                }
                JSONStorableAction rHbuttonAct = new JSONStorableAction("Hand:Right to Left", () =>{CopyHandFromRightToLeft();});					
                RegisterAction(rHbuttonAct);

                rFbutton = CreateButton("Foot : Right to Left", true);
                if (rFbutton != null)
                {
                    rFbutton.button.onClick.AddListener(CopyFootFromRightToLeft);
                }
                JSONStorableAction rFbuttonAct = new JSONStorableAction("Foot:Right to Left", () =>{CopyFootFromRightToLeft();});					
                RegisterAction(rFbuttonAct);
                
                ubutton = CreateButton("Undo");
                if (ubutton != null)
                {
                    ubutton.button.onClick.AddListener(Undo);
                }
                JSONStorableAction ubuttonAct = new JSONStorableAction("Undo", () =>{Undo();});					
                RegisterAction(ubuttonAct);
                #endregion

                PosiST["On"] = FreeControllerV3.PositionState.On;
                PosiST["Comply"] = FreeControllerV3.PositionState.Comply;
                PosiST["Off"] = FreeControllerV3.PositionState.Off;
                PosiST["ParentLink"] = FreeControllerV3.PositionState.ParentLink;
                PosiST["PhysicsLink"] = FreeControllerV3.PositionState.PhysicsLink;
                PosiST["Hold"] = FreeControllerV3.PositionState.Hold;
                PosiST["Lock"] = FreeControllerV3.PositionState.Lock;

                RotaST["On"] = FreeControllerV3.RotationState.On;
                RotaST["Comply"] = FreeControllerV3.RotationState.Comply;
                RotaST["Off"] = FreeControllerV3.RotationState.Off;
                RotaST["ParentLink"] = FreeControllerV3.RotationState.ParentLink;
                RotaST["PhysicsLink"] = FreeControllerV3.RotationState.PhysicsLink;
                RotaST["Hold"] = FreeControllerV3.RotationState.Hold;
                RotaST["Lock"] = FreeControllerV3.RotationState.Lock;
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        void Start()
        {
            UpdateInitialMorphs();
        }
        
        private void clearUndo(string dummy) {
            canUndo = false;
        }

        public void mkList() {
            List<string> targetChoices = new List<string>();
            foreach (string atomUID in SuperController.singleton.GetAtomUIDs()) {
                var targetAtom = SuperController.singleton.GetAtomByUid(atomUID);
                if (targetAtom.type == "Person") {
                    targetChoices.Add(atomUID);
                } 
            }                            
			atomJSON.choices = targetChoices;
            if (containingAtom.type == "Person") {
                atomJSON.val = containingAtom.name;
                CurrentAtom = SuperController.singleton.GetAtomByUid(atomJSON.val);
            } else {
			    atomJSON.val = "None";
            }
        }

        private void CopyHandFromLeftToRight()
        {
            HorF = "Hand";
            CF = "Left ";
            CF1 = "l";
            CT = "Right ";
            CT1 = "r";
            Mirroring();
        }

        private void CopyFootFromLeftToRight()
        {
            HorF = "Foot";
            CF = "Left ";
            CF1 = "l";
            CT = "Right ";
            CT1 = "r";
            Mirroring();
        }

        private void CopyHandFromRightToLeft()
        {
            HorF = "Hand";
            CF = "Right ";
            CF1 = "r";
            CT = "Left ";
            CT1 = "l";
            Mirroring();
        }

        private void CopyFootFromRightToLeft()
        {
            HorF = "Foot";
            CF = "Right ";
            CF1 = "r";
            CT = "Left ";
            CT1 = "l";
            Mirroring();
        }

        private void UpdateInitialMorphs()
        {
            JSONStorable geometry = CurrentAtom.GetStorableByID("geometry");
            DAZCharacterSelector character = geometry as DAZCharacterSelector;
            GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
        }

        private void ApplyMorphs()
        {
            JSONStorable geometry = CurrentAtom.GetStorableByID("geometry");
            DAZCharacterSelector character = geometry as DAZCharacterSelector;
            GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;

            List<string> lrPoseList = new List<string>(lrPoseNames);
            foreach (string lrPoseNames in lrPoseList)
            {
                float tempPose_r = morphControl.GetMorphByDisplayName("Right " + lrPoseNames).morphValue; 
                float tempPose_l = morphControl.GetMorphByDisplayName("Left " + lrPoseNames).morphValue;
                HandPose["Right " + lrPoseNames] = tempPose_r;
                HandPose["Left " + lrPoseNames] = tempPose_l;
            }

            if (HorF == "Hand")
            {
                List<string> lrHandPoseList = new List<string>(lrHandPoseNames);
                foreach (string lrHandPoseNames in lrHandPoseList)
                {
                    morphControl.GetMorphByDisplayName(CT + lrHandPoseNames).morphValue = HandPose[CF + lrHandPoseNames];
                }
            }
            else
            {
                List<string> lrFootPoseList = new List<string>(lrFootPoseNames);
                foreach (string lrFootPoseNames in lrFootPoseList)
                {
                    morphControl.GetMorphByDisplayName(CT + lrFootPoseNames).morphValue = HandPose[CF + lrFootPoseNames]; 
                }
            }
        }

        public void Mirroring()
        {
            if (atomJSON.val != "None")
            {
                CurrentAtom = SuperController.singleton.GetAtomByUid(atomJSON.val);
                canUndo = true;
                Copy();
                Invoke("StatusRecall", 0.5f);
                Invoke("ColButtonCallback", 0.5f);
            }
        }

        public void Copy()
        {
            CurrentAtom.collisionEnabledJSON.val=false;

            Rigidbody[] objects = CurrentAtom.rigidbodies;
            foreach (var ObjectName in objects)
            {
                RigidbodyList[ObjectName.ToString()] = ObjectName;
            }
            
            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                PosiDic[ControlNames] = CurrentController.currentPositionState.ToString();
                RotaDic[ControlNames] = CurrentController.currentRotationState.ToString();

                if (!CurrentController.linkToRB)
                {
                    AtomDic[ControlNames] = "None";
                    LinkDic[ControlNames] = "None"; 
                    RigidbodyList["None"] = CurrentController.linkToRB;
                }
                else
                {
                    var jc = CurrentController.GetJSON(); 
                    var linkTo = jc["linkTo"].Value;
                    AtomDic[ControlNames] = linkTo.Split(":"[0])[0];
                    LinkDic[ControlNames] = CurrentController.linkToRB.ToString();
                }

                CurrentController.SetLinkToAtom("None");
                CurrentController.SetLinkToRigidbodyObject("None");
                CurrentController.currentPositionState = PosiST["On"];
                CurrentController.currentRotationState = RotaST["On"];
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                PosiDic["r" + lrControlNames] = rCurrentController.currentPositionState.ToString();
                RotaDic["r" + lrControlNames] = rCurrentController.currentRotationState.ToString();

                if (!rCurrentController.linkToRB)
                {
                    AtomDic["r" + lrControlNames] = "None";
                    LinkDic["r" + lrControlNames] = "None";
                }
                else
                {
                    var jc = rCurrentController.GetJSON(); 
                    var linkTo = jc["linkTo"].Value;
                    AtomDic["r" + lrControlNames] = linkTo.Split(":"[0])[0];
                    LinkDic["r" + lrControlNames] = rCurrentController.linkToRB.ToString();
                }

                rCurrentController.SetLinkToAtom("None");
                rCurrentController.SetLinkToRigidbodyObject("None");
                rCurrentController.currentPositionState = PosiST["On"];
                rCurrentController.currentRotationState = RotaST["On"];

                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                PosiDic["l" + lrControlNames] = lCurrentController.currentPositionState.ToString();
                RotaDic["l" + lrControlNames] = lCurrentController.currentRotationState.ToString();

                if (!lCurrentController.linkToRB)
                {
                    AtomDic["l" + lrControlNames] = "None";
                    LinkDic["l" + lrControlNames] = "None"; 
                }
                else
                {
                    var jc = lCurrentController.GetJSON(); 
                    var linkTo = jc["linkTo"].Value;
                    AtomDic["l" + lrControlNames] = linkTo.Split(":"[0])[0];
                    LinkDic["l" + lrControlNames] = lCurrentController.linkToRB.ToString();
                }

                lCurrentController.SetLinkToAtom("None");
                lCurrentController.SetLinkToRigidbodyObject("None");
                lCurrentController.currentPositionState = PosiST["On"];
                lCurrentController.currentRotationState = RotaST["On"];
            }

            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;

                if (ControlNames == "chestControl") 
                {
                    tempPosition_h.x=CurrentController.transform.localPosition.x;
                }

                PosiV3[ControlNames] = CurrentController.transform.localPosition;
                RotaV3[ControlNames] = CurrentController.transform.localRotation;
            }

            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                PosiV3["l" + lrControlNames] = lCurrentController.transform.localPosition;
                RotaV3["l" + lrControlNames] = lCurrentController.transform.localRotation;

                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                PosiV3["r" + lrControlNames] = rCurrentController.transform.localPosition;
                RotaV3["r" + lrControlNames] = rCurrentController.transform.localRotation;
            }

            List<string> lrHandControlList = new List<string>(lrHandControlNames);
            List<string> lrFootControlList = new List<string>(lrFootControlNames);
            if (HorF == "Hand")
            {
                foreach (string lrHandControlNames in lrHandControlList)
                {
                    temprPosition.x = CurrentAtom.transform.localPosition.x - PosiV3[CF1 + lrHandControlNames].x + tempPosition_h.x * 2;
                    temprPosition.y = PosiV3[CF1 + lrHandControlNames].y;
                    temprPosition.z = PosiV3[CF1 + lrHandControlNames].z;
                    temprRotation.x = RotaV3[CF1 + lrHandControlNames].x;
                    temprRotation.y = CurrentAtom.transform.localRotation.y - RotaV3[CF1 + lrHandControlNames].y;
                    temprRotation.z = CurrentAtom.transform.localRotation.z - RotaV3[CF1 + lrHandControlNames].z;
                    temprRotation.w = RotaV3[CF1 + lrHandControlNames].w;

                    FreeControllerV3 toCurrentHandController = CurrentAtom.GetStorableByID(CT1 + lrHandControlNames) as FreeControllerV3;
                    toCurrentHandController.transform.localPosition = temprPosition;
                    toCurrentHandController.transform.localRotation = temprRotation;
                }
            }
            else
            {
                foreach (string lrFootControlNames in lrFootControlList)
                {
                    temprPosition.x = CurrentAtom.transform.localPosition.x - PosiV3[CF1 + lrFootControlNames].x + tempPosition_h.x * 2;
                    temprPosition.y = PosiV3[CF1 + lrFootControlNames].y;
                    temprPosition.z = PosiV3[CF1 + lrFootControlNames].z;
                    temprRotation.x = RotaV3[CF1 + lrFootControlNames].x;
                    temprRotation.y = CurrentAtom.transform.localRotation.y - RotaV3[CF1 + lrFootControlNames].y;
                    temprRotation.z = CurrentAtom.transform.localRotation.z - RotaV3[CF1 + lrFootControlNames].z;
                    temprRotation.w = RotaV3[CF1 + lrFootControlNames].w;

                    FreeControllerV3 toCurrentFootController = CurrentAtom.GetStorableByID(CT1 + lrFootControlNames) as FreeControllerV3;
                    toCurrentFootController.transform.localPosition = temprPosition;
                    toCurrentFootController.transform.localRotation = temprRotation;
                }
            }
            ApplyMorphs();
        }

        public void StatusRecall()
        {
            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                CurrentController.currentPositionState = PosiST[PosiDic[ControlNames]];
                CurrentController.currentRotationState = RotaST[RotaDic[ControlNames]];

                if (AtomDic[ControlNames] != "None")
                {
                    RigidbodyName = LinkDic[ControlNames];
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[ControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                    CurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }                
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                rCurrentController.SetLinkToAtom(AtomDic["r" + lrControlNames]);
                lCurrentController.SetLinkToAtom(AtomDic["l" + lrControlNames]);
                rCurrentController.currentPositionState = PosiST[PosiDic["r" + lrControlNames]];
                rCurrentController.currentRotationState = RotaST[RotaDic["r" + lrControlNames]];
                lCurrentController.currentPositionState = PosiST[PosiDic["l" + lrControlNames]];
                lCurrentController.currentRotationState = RotaST[RotaDic["l" + lrControlNames]];

                if (LinkDic["l" + lrControlNames] != "None")
                {
                    RigidbodyName = LinkDic["l" + lrControlNames];
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic["l" + lrControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                    lCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }
                else
                {
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    lCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }

                if (LinkDic["r" + lrControlNames] != "None")
                {
                    RigidbodyName = LinkDic["r" + lrControlNames];
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic["r" + lrControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                    rCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }
                else
                {
                    linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    rCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }
            }

            List<string> lrHandControlList = new List<string>(lrHandControlNames);
            List<string> lrFootControlList = new List<string>(lrFootControlNames);
            if (HorF == "Hand")
            {
                foreach (string lrHandControlNames in lrHandControlList)
                {
                    FreeControllerV3 toCurrentHandController = CurrentAtom.GetStorableByID(CT1 + lrHandControlNames) as FreeControllerV3;
                    toCurrentHandController.SetLinkToAtom(AtomDic[CF1 + lrHandControlNames]);
                    toCurrentHandController.currentPositionState = PosiST[PosiDic[CF1 + lrHandControlNames]];
                    toCurrentHandController.currentRotationState = RotaST[RotaDic[CF1 + lrHandControlNames]];

                    if (LinkDic[CF1 + lrHandControlNames] != "None")
                    {
                        RigidbodyName = LinkDic[CF1 + lrHandControlNames];
                        if (RigidbodyName.Substring(0, 1) == "l")
                        {
                            RigidbodyName = "r" + RigidbodyName.Substring(1, RigidbodyName.Length - 1);
                        }
                        else if (RigidbodyName.Substring(0, 1) == "r")
                        {
                            RigidbodyName = "l" + RigidbodyName.Substring(1, RigidbodyName.Length - 1);
                        }
                        linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[CF1 + lrHandControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                    }
                    else
                    {
                        linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    }
                    toCurrentHandController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }
            }
            else
            {
                foreach (string lrFootControlNames in lrFootControlList)
                {
                    FreeControllerV3 toCurrentFootController = CurrentAtom.GetStorableByID(CT1 + lrFootControlNames) as FreeControllerV3;
                    toCurrentFootController.SetLinkToAtom(AtomDic[CF1 + lrFootControlNames]);
                    toCurrentFootController.currentPositionState = PosiST[PosiDic[CF1 + lrFootControlNames]];
                    toCurrentFootController.currentRotationState = RotaST[RotaDic[CF1 + lrFootControlNames]];

                    if (LinkDic[CF1 + lrFootControlNames] != "None")
                    {
                        RigidbodyName = LinkDic[CF1 + lrFootControlNames];
                        if (RigidbodyName.Substring(0, 1) == "l")
                        {
                            RigidbodyName = "r" + RigidbodyName.Substring(1, RigidbodyName.Length - 1);
                        }
                        else if (RigidbodyName.Substring(0, 1) == "r")
                        {
                            RigidbodyName = "l" + RigidbodyName.Substring(1, RigidbodyName.Length - 1);
                        }
                        linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[CF1 + lrFootControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                    }
                    else
                    {
                        linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody("None:None");
                    }
                    toCurrentFootController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
                }
            }
        }

        public void ColButtonCallback() 
        {
            CurrentAtom.collisionEnabledJSON.val = true;
        }

        public void Undo()
        {
            if (atomJSON.val != "None" && canUndo == true)
            {
                CurrentAtom = SuperController.singleton.GetAtomByUid(atomJSON.val);
                AllControlOn();
                UndoMain();
                Invoke("StatusRecall2", 0.5f);
                Invoke("ColButtonCallback", 0.5f);
            }
        }

        public void AllControlOn()
        {
            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                CurrentController.currentPositionState = PosiST["On"];
                CurrentController.currentRotationState = RotaST["On"];
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                rCurrentController.currentPositionState = PosiST["On"];
                rCurrentController.currentRotationState = RotaST["On"];
                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                lCurrentController.currentPositionState = PosiST["On"];
                lCurrentController.currentRotationState = RotaST["On"];
            }
        }

        public void StatusRecall2()
        {
            Rigidbody[] objects = CurrentAtom.rigidbodies;
            foreach (var ObjectName in objects)
            {
                RigidbodyList[ObjectName.ToString()] = ObjectName;
            }
            
            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                CurrentController.currentPositionState = PosiST[PosiDic[ControlNames]];
                CurrentController.currentRotationState = RotaST[RotaDic[ControlNames]];
                RigidbodyName = LinkDic[ControlNames];
                linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic[ControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                CurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                rCurrentController.currentPositionState = PosiST[PosiDic["r" + lrControlNames]];
                rCurrentController.currentRotationState = RotaST[RotaDic["r" + lrControlNames]];
                RigidbodyName = LinkDic["r" + lrControlNames];
                linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic["r" + lrControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                rCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);

                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                lCurrentController.currentPositionState = PosiST[PosiDic["l" + lrControlNames]];
                lCurrentController.currentRotationState = RotaST[RotaDic["l" + lrControlNames]];
                RigidbodyName = LinkDic["l" + lrControlNames];
                linkToAtom = SuperController.singleton.RigidbodyNameToRigidbody(AtomDic["l" + lrControlNames] + ":" + RigidbodyName.ToString().Split(" "[0])[0]);
                lCurrentController.SelectLinkToRigidbody(linkToAtom, FreeControllerV3.SelectLinkState.PositionAndRotation);
            }
        }

        public void UndoMain()
        {
            CurrentAtom.collisionEnabledJSON.val=false;

            Rigidbody[] objects = CurrentAtom.rigidbodies;
            foreach (var ObjectName in objects)
            {
                RigidbodyList[ObjectName.ToString()] = ObjectName;
            }
            
            List<string> ControlList = new List<string>(ControlNames);
            foreach (string ControlNames in ControlList)
            {
                FreeControllerV3 CurrentController = CurrentAtom.GetStorableByID(ControlNames) as FreeControllerV3;
                CurrentController.transform.localPosition = PosiV3[ControlNames];
                CurrentController.transform.localRotation = RotaV3[ControlNames];
            }

            List<string> lrControlList = new List<string>(lrControlNames);
            foreach (string lrControlNames in lrControlList)
            {
                FreeControllerV3 rCurrentController = CurrentAtom.GetStorableByID("r" + lrControlNames) as FreeControllerV3;
                rCurrentController.transform.localPosition = PosiV3["r" + lrControlNames];
                rCurrentController.transform.localRotation = RotaV3["r" + lrControlNames];

                FreeControllerV3 lCurrentController = CurrentAtom.GetStorableByID("l" + lrControlNames) as FreeControllerV3;
                lCurrentController.transform.localPosition = PosiV3["l" + lrControlNames];
                lCurrentController.transform.localRotation = RotaV3["l" + lrControlNames];
            }

            JSONStorable geometry = CurrentAtom.GetStorableByID("geometry");
            DAZCharacterSelector character = geometry as DAZCharacterSelector;
            GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;

            List<string> lrPoseList = new List<string>(lrPoseNames);
            foreach (string lrPoseNames in lrPoseList)
            {
                morphControl.GetMorphByDisplayName("Right " + lrPoseNames).morphValue = HandPose["Right " + lrPoseNames]; 
                morphControl.GetMorphByDisplayName("Left " + lrPoseNames).morphValue = HandPose["Left " + lrPoseNames];
            }
        }
    }
}

